# (c) Copyright 2005-2010. CodeWeavers, Inc.
package CXMenu;
use warnings;
use strict;
use CXLog;
use CXUtils;


#####
#
# Miscellaneous
#
#####

sub get_default_icon($)
{
    my ($is_dir)=@_;
    return CXUtils::get_std_icon($is_dir ? "cxmenu.xpm" : "crossover.xpm");
}

sub get_icon_path($$)
{
    my ($wineprefix, $path)=@_;

    return "$wineprefix/windata/" . mangle_path("$path.xpm");
}

sub menu_to_lnk($)
{
    my ($menu)=@_;
    if ($menu =~ s!^StartMenu/!c:/Windows/Start Menu/!)
    {
        # Nothing to do
    }
    elsif ($menu =~ s!^StartMenu([._])([^/]+)/!!)
    {
        $menu=join("", demangle_string($2),
                       ($1 eq "." ? "/" : "/Start Menu/"),
                       $menu);
    }
    elsif ($menu =~ s!^Desktop/!c:/Windows/Desktop/!)
    {
        # Nothing to do
    }
    elsif ($menu =~ s!^Desktop([._])([^/]+)/!!)
    {
        $menu=join("", demangle_string($2),
                       ($1 eq "." ? "/" : "/Desktop/"),
                       $menu);
    }
    else
    {
        cxerr("unknown menu location '$menu'\n");
    }
    return $menu;
}

sub set_path_closure($$$)
{
    my ($hash, $path, $value)=@_;

    # We don't want any trailing slash
    delete $hash->{$path} if ($path =~ m%/$%);

    while (1)
    {
        last if ($path !~ s%/+[^/]*$%%);
        last if (exists $hash->{$path});
        $hash->{$path}=$value;
    }
}

sub wineshelllink2cxmenu($$$$$)
{
    my ($ismenu, $root, $link, $path, $icon)=@_;
    my $cxmenu;
    $@="";

    # Convert the .lnk path into a suitable menu name
    my $menu=($ismenu ? "StartMenu" : "Desktop");
    $root =~ s!\\+!/!g;
    if ($root !~ m!^c:/Windows/(Start Menu|Desktop)!i)
    {
        # Keep the exact location of the .lnk file,
        # otherwise we will not be able to run it later
        $menu .= "." . mangle_string($root);
    }
    $link =~ s!\\+!/!g;
    if ($link !~ s!^\Q$root/\E!/!i)
    {
        $@="the link location '$link' does not match the root location '$root'\n";
        return undef;
    }
    $menu.=$link;
    $cxmenu->{fullpath}=$menu;

    if ($icon)
    {
        # Move the icon to its final resting place. Failure to do this is a
        # warning, not a fatal error.
        $cxmenu->{icon}=CXMenu::get_icon_path($ENV{WINEPREFIX}, $menu);
        my $icon_dir=cxdirname($cxmenu->{icon});
        if (!-d $icon_dir and !cxmkpath($icon_dir))
        {
            $@="unable to create '$icon_dir': $@\n";
        }
        elsif (!cxmv($icon, $cxmenu->{icon}))
        {
            $@="unable to move '$icon' to '$cxmenu->{icon}': $!\n";
        }
    }

    if ($path =~ s/\.exe$//i)
    {
        $path =~ s%^.*[/\\]%%;
        $path =~ s/ //g;
        $path =~ tr/A-Z/a-z/;
        # Blacklist some shortcut names
        if ($path !~ /(?:^(?:autorun|start|unins|unvise)|~)/i and
            ($path ne "winword" or $menu !~ /iTunes/i) and
            ($path ne "winproj" or $menu !~ /Server Accounts/i))
        {
            $cxmenu->{shortcut}=$path;
        }
    }
    return $cxmenu;
}


#####
#
# KDE / Gnome locale and encoding handling
#
#####

# In Gnome < 2.0, the .desktop files use the 'legacy-mixed' encoding,
# i.e. a locale-dependent encoding defined by the following table.
my $gnome_encoding_to_locale={
    "BIG5"        => ["zh_TW"],
    "CP1251"      => ["be","bg"],
    "EUC-CN"      => ["zh_CN"],
    "EUC-JP"      => ["ja"],
    "EUC-KR"      => ["ko"],
    "ISO-8859-1"  => ["br","ca","da","de","en","es","eu","fi","fr","gl",
                      "it","nl","no","pt","sv","wa"],
    "ISO-8859-2"  => ["cs","hr","hu","pl","ro","sk","sl","sq","sr"],
    "ISO-8859-3"  => ["eo"],
    "ISO-8859-5"  => ["mk","sp"],
    "ISO-8859-7"  => ["el"],
    "ISO-8859-9"  => ["tr"],
    "ISO-8859-13" => ["lt","lv","mi"],
    "ISO-8859-14" => ["cy","ga"],
    "ISO-8859-15" => ["et"],
    "KOI8-R"      => ["ru"],
    "KOI8-U"      => ["uk"],
    "TIS-620"     => ["th"]
};

my $gnome_locale_to_encoding;
sub gnome_locale_to_encoding($)
{
    if (!$gnome_locale_to_encoding)
    {
        while (my ($encoding, $locales)=each %$gnome_encoding_to_locale)
        {
            map { $gnome_locale_to_encoding->{$_}=$encoding; } @$locales;
        }
    }

    my ($locale)=@_;
    # FIXME: If the (system) locale specifies an encoding, should we use it?
    $locale =~ s/[@.].*$//;
    my $encoding=$gnome_locale_to_encoding->{$locale};
    if (!defined $encoding)
    {
        $locale =~ s/_.*$//;
        $encoding=$gnome_locale_to_encoding->{$locale};
    }
    $encoding="ISO-8859-1" if (!defined $encoding);

    return $encoding;
}

sub gnome_cxgettext_setup($$)
{
    my ($locale, $encoding)=@_;

    CXUtils::cxsetlang($locale);
    if ($encoding eq "legacy-mixed")
    {
        CXUtils::cxsetencoding(gnome_locale_to_encoding($locale));
    }
}


#####
#
# KDE / Gnome / VFolder .desktop files
#
#####

sub kg_fill_desktop_entry($$$)
{
    my ($section, $component, $tags)=@_;

    if ($component->{_encoding} ne "legacy-mixed")
    {
        $section->set("Encoding", $component->{_encoding});
    }

    $section->set("Type", ($component->{is_dir} ? "Directory" : "Application"));
    $section->set("X-Created-By", $tags);
    $section->set("Categories", $component->{_categories}) if (defined $component->{_categories});

    my ($icon, $exec, $name, $comment);
    if ($component->{_encoding} ne "UTF-8")
    {
        my $locale=$ENV{LC_ALL} || $ENV{LC_MESSAGES} || $ENV{LANG} || "en";
        my $encoding=gnome_locale_to_encoding($locale);
        require CXRecode;
        $icon=CXRecode::recode($component->{icon}, "UTF-8", $encoding);
        $exec=CXRecode::recode($component->{command}, "UTF-8", $encoding);
        $name=CXRecode::recode($component->{name}, "UTF-8", $encoding);
        $comment=CXRecode::recode($component->{description}, "UTF-8", $encoding);
    }
    else
    {
        $icon=$component->{icon};
        $exec=$component->{command};
        $name=$component->{name};
        $comment=$component->{description};
    }
    $section->set("Icon", $icon);
    if ($exec)
    {
        # Notes:
        # * The desktop file format specifies that percents must be doubled.
        #   Gnome 1.x is an exception but does not handle them in a consistent
        #   way anyway.
        # * Not specified there is that KDE is buggy and is unable to properly
        #   handle double quotes even if we try to escape them.
        # * The XDG desktop file format also specifies that escape sequences
        #   are supported and that backslashes must be escaped.
        $exec =~ s/%/%%/g;
        $section->set("Exec", $exec);
    }
    $section->set("Name", "$name$component->{_creator}");
    $section->set("Comment", $comment) if ($comment);

    if ($component->{localize})
    {
        # Translate the name and description in two passes
        # so the file looks nice.
        my $oldlang=CXUtils::cxgetlang();
        my $oldencoding=CXUtils::cxsetencoding($component->{_encoding});
        foreach my $locale (CXUtils::get_supported_locales())
        {
            gnome_cxgettext_setup($locale, $component->{_encoding});
            my $msgstr=cxgettext($component->{name});
            if ($msgstr ne "" and $msgstr ne $component->{name})
            {
                $section->set("Name[$locale]", "$msgstr$component->{_creator}");
            }
        }
        if ($component->{description})
        {
            foreach my $locale (CXUtils::get_supported_locales())
            {
                gnome_cxgettext_setup($locale, $component->{_encoding});
                my $msgstr=cxgettext($component->{description});
                if ($msgstr ne "" and $msgstr ne $component->{description})
                {
                    $section->set("Comment[$locale]", $msgstr);
                }
            }
        }
        CXUtils::cxsetlang($oldlang);
        CXUtils::cxsetencoding($oldencoding);
    }
}

sub kg_create_desktop_file($$)
{
    my ($filename, $component)=@_;

    # Create the desktop file from scratch
    require CXRWConfig;
    my $desktop=CXRWConfig->new(undef, $component->{_escaping}, "");
    cxlog("Creating '$filename'\n");
    $desktop->set_filename($filename);

    my $section=$desktop->append_section($component->{_section_name});
    kg_fill_desktop_entry($section, $component, $component->{tag});
    $desktop->set_shebang($component->{_shebang}) if (defined $component->{_shebang});

    if (!$desktop->save())
    {
        cxerr("unable to save '$filename': $!\n");
        return undef;
    }
    chmod(0777 & ~umask(), $filename) if (defined $component->{_shebang});
    return $desktop;
}


#####
#
# KDE / Gnome / VFolder .directory files
#
#####

sub kg_create_directory($$$)
{
    my ($filename, $component, $save)=@_;

    cxlog("Creating '$filename'\n");
    require CXRWConfig;
    my $directory=CXRWConfig->new($filename, $component->{_escaping}, "");
    my $section=$directory->get_section("Desktop Entry");
    my %tags;
    if ($section)
    {
        my $created_by=$section->get("X-Created-By");

        # Don't tag non-CrossOver menus
        return $directory if (!defined $created_by);

        # Recreate the directory from scratch, only preserve the
        # X-Created-By field
        map { $tags{$_}=1; } split /;+/, $created_by;
        if (!$component->{intermediate})
        {
            $directory->remove_all();
            $section=undef;
        }
    }
    elsif (!$component->{intermediate})
    {
        # Reset the file, it's ours now
        $directory->remove_all();
    }
    elsif (-f "$filename")
    {
        cxerr("unknown format for '$filename'\n");
        return undef;
    }
    $tags{$component->{tag}}=1;
    my $created_by=join(";", sort keys %tags);

    if (!$section)
    {
        $section=$directory->append_section("Desktop Entry");
        kg_fill_desktop_entry($section, $component, $created_by);
    }
    else
    {
        $section->set("X-Created-By", $created_by);
    }
    if ($save and !$directory->save())
    {
        cxerr("unable to save '$filename': $!\n");
    }
    return $directory;
}


#####
#
# The CXMenu base class
#
#####


# When creating menus, intermediate folders are created implicitly. This
# functions is responsible for 'garbage collecting' them when they become
# empty.
#
# In order to call this function a class must setup a list of folder paths
# to be garbage collected in $self->{garbage_collect} and a list of folder
# path status in $self->{status}.
#
# It must also implement the following helper functions:
# * ($status, $data)=$self->gc_get_status($path);
# * $self->gc_delete($path, $data)
# * $self->gc_untag($path, $data)
#
# The status can be one of the following:
# * empty
#   The folder does not contain any sub-folder or menu and thus can be deleted
#
# * non-empty
#   The folder still contains menus or sub-folders although none of them
#   belong to the current bottle. So we can remove the bottle's tag but we
#   must not delete the folder.
#
# * in-use
#   The folder contains at least one menu or sub-folder belonging to the
#   current bottle. This means it must not be deleted and that the bottle's
#   tag must not be removed either.
#
# * alien
#   This folder does not belong to us at all. There is no tag to remove (by
#   definition) and we should not delete the folder.
sub garbage_collect($)
{
    my ($self)=@_;

    my $garbage_collect=$self->{garbage_collect};
    return if (!$garbage_collect or !%$garbage_collect);

    # If a folder is marked for garbage collection,
    # then all its parents should be too
    foreach my $path (sort { $b cmp $a } keys %$garbage_collect)
    {
        CXMenu::set_path_closure($garbage_collect, $path, 1);
    }
    # For debug:
    # cxlog("Initial garbage collection list:\n");
    # map { cxlog("  $_\n") } sort { $b cmp $a } keys %$garbage_collect;

    # If a folder is 'in-use', then all its parents are 'in-use' too
    # In fact the more correct status may be 'alien' but it does not
    # matter because we won't touch them in either case
    my $status=$self->{status};
    foreach my $path (sort { $b cmp $a } keys %$status)
    {
        CXMenu::set_path_closure($status, $path, "in-use");
    }
    # For debug:
    # cxlog("Initial status:\n");
    # map { cxlog("  $_ -> in-use\n") } sort { $b cmp $a } keys %$status;

    # Now do the garbage collection itself
    foreach my $path (sort { $b cmp $a } keys %$garbage_collect)
    {
        cxlog("garbage_collect: considering '$path'\n");
        if (exists $status->{$path})
        {
            # If a folder already has a status it is necessarily
            # 'in-use' which means we should not touch it
            cxlog(" -> already $status->{$path}\n");
            next;
        }
        my ($st, $data)=$self->gc_get_folder_status($path);
        next if (!defined $st);
        $status->{$path}=$st;
        cxlog(" -> $st\n");
        if ($st eq "empty")
        {
            # Delete the folder altogether
            $self->gc_delete_folder($path, $data);
        }
        elsif ($st eq "non-empty")
        {
            # Just untag the folder
            $self->gc_untag_folder($path, $data);
        }
        elsif ($st eq "in-use")
        {
            CXMenu::set_path_closure($status, $path, "in-use");
        }
    }
}

sub finalize($)
{
    my ($self)=@_;
    $self->garbage_collect();
    return 1;
}

return 1;
