# (c) Copyright 2002-2010. CodeWeavers, Inc.
package CXMenuVFolder;
use warnings;
use strict;

use CXLog;
use CXUtils;
use CXXMLDOM;
use CXMenu;
use base "CXMenu";


#####
#
# Detect the VFolder secondary directories
#
#####

sub uses_parent_tags($)
{
    my ($self)=@_;

    if (!defined $self->{parent_tags})
    {
        my $node=get_child($self->{menu_root}, "Parent");
        $self->{parent_tags}=(defined $node);
    }
    return $self->{parent_tags};
}

sub get_write_dir($)
{
    my ($self)=@_;

    return $self->{write_dir} if (defined $self->{write_dir});

    # Try to detect the WriteDir
    my $other_dirs;
    if ($> != 0 or $self->{menu} eq $self->{managed_menu})
    {
        $other_dirs=$self->{other_dirs};
    }

    my $write_dir;
    foreach my $dir (@{$self->{write_dirs}}, @$other_dirs,
                     cxdirname($self->{menu}) . "/applications")
    {
        next if (!defined $dir);
        if ((-d $dir and -w $dir) or (!-d $dir and cxmkpath($dir)))
        {
            cxlog("Using write_dir='$dir'\n");
            $write_dir=$dir;
            last;
        }
        cxlog("Ignoring potential write_dir '$dir'\n");
    }

    if (defined $write_dir)
    {
        # Also make sure we have an ItemDir or MergeDir pointing to it,
        # otherwise VFolders will ignore it
        if (!grep /^\Q$write_dir\E\/?$/, @{$self->{other_dirs}})
        {
            my $tag;
            if (CXUtils::file_grep("/etc/SuSE-release", "^VERSION\\s+=\\s+8\\.1\\b"))
            {
                # On SuSE 8.1 ItemDir causes all the menus to disappear so we
                # use MergeDir instead. Note that this is really specific to
                # SuSE 8.1, not to Gnome 2.0.5.
                $tag="MergeDir";
            }
            else
            {
                # Otherwise ItemDir is the right tag for this. Besides,
                # MergeDir can cause menu duplication.
                $tag="ItemDir";
            }
            cxlog("Adding a $tag tag for WriteDir\n");
            add_new_element($self->{vfolderinfo}, $tag, $write_dir,
                               $self->{vfolderinfo}->getFirstChild());
            push @{$self->{write_dirs}}, $write_dir;
        }
        $self->{write_dir}=$write_dir;
    }
    else
    {
        cxlog("Could not find a suitable write_dir!\n");
    }
    return $write_dir;
}

sub get_desktop_dir($)
{
    my ($self)=@_;

    return $self->{desktop_dir} if (defined $self->{desktop_dir});

    # Is the DesktopDir specified in the XML file?
    my $desktop_dir;
    foreach my $dir (@{$self->{desktop_dirs}}, @{$self->{write_dirs}})
    {
        if (-d $dir and -w $dir)
        {
            $desktop_dir=$dir;
            cxlog("Setting desktop_dir='$dir'\n");
            last;
        }
    }

    # If not, try to guess it anyway
    if (!defined $desktop_dir)
    {
        if (!$self->uses_parent_tags() and
            $self->{menu} ne $self->{managed_menu})
        {
            # This is a Gnome 2.0 system -> we must put .directory files in
            # the same directory as the XML file
            $desktop_dir=cxdirname($self->{menu});
        }
        else
        {
            $desktop_dir=$self->get_write_dir();
        }
        if (defined $desktop_dir)
        {
            cxlog("Using desktop_dir='$desktop_dir'\n");
        }
        else
        {
            cxlog("Could not find a suitable desktop_dir!\n");
        }
    }
    $self->{desktop_dir}=$desktop_dir;
    return $desktop_dir;
}


#####
#
# The VFolder Cache
# The vfc_xxx functions read CrossOver .desktop and .directory files
# found in the many VFolder folders and cache for quick retrieval by
# basename and by category.
#
#####

sub vfc_add_file($$)
{
    my ($self, $file)=@_;

    # If the cache has not been initialized yet then just ignore this request
    return if (!defined $self->{vfc_files});

    my $filename=$file->get_filename();
    my $basename=$filename;
    $basename=~ s%^.*/%%;
    $self->{vfc_files}->{$basename}=$file;
    my $tags=$file->get("Desktop Entry", "X-Created-By", "");
    foreach my $tag (split /;+/, $tags)
    {
        $self->{vfc_tags}->{$tag}->{$filename}=$file;
        $file->{_vfc_tags}->{$tag}=1;
    }
    my $categories=$file->get("Desktop Entry", "Categories", "");
    foreach my $category (split /;+/, $categories)
    {
        $self->{vfc_categories}->{$category}->{$filename}=$file;
        $file->{_vfc_categories}->{$category}=1;
    }
}

sub vfc_init()
{
    my ($self)=@_;

    $self->{vfc_files}={};
    $self->{vfc_tags}={};
    $self->{vfc_categories}={};
    my $write_extra=cxdirname($self->{menu});
    foreach my $dir (@{$self->{write_dirs}}, @{$self->{desktop_dirs}},
                     @{$self->{other_dirs}}, $write_extra,
                     "$write_extra/applications")
    {
        next if (!-d $dir);
        if (!-w $dir)
        {
            cxlog("skipping read-only directory '$dir'\n");
            next;
        }
        my $dh;
        if (!opendir($dh, $dir))
        {
            cxlog("unable to open the '$dir' directory: $!\n");
            next;
        }

        require CXRWConfig;
        foreach my $dentry (readdir $dh)
        {
            next if ($dentry !~ /^CXTree_.*\.(?:desktop|directory)$/);
            $dentry="$dir/$dentry";
            my $file=CXRWConfig->new($dentry, undef, "");
            if (!$file)
            {
                cxlog("unable to read '$dentry'\n");
                next;
            }
            $self->vfc_add_file($file);
        }
        closedir($dh);
    }
}

sub vfc_get_file($$)
{
    my ($self, $basename)=@_;

    if (defined $self->{vfc_files})
    {
        return $self->{vfc_files}->{$basename};
    }

    # It would be overkill to read all the files just because we need one
    # So just search for this one file, and set it up
    my $write_extra=cxdirname($self->{menu});
    foreach my $dir (@{$self->{write_dirs}}, @{$self->{desktop_dirs}},
                     @{$self->{other_dirs}}, $write_extra,
                     "$write_extra/applications")
    {
        if (-f "$dir/$basename")
        {
            require CXRWConfig;
            my $file=CXRWConfig->new("$dir/$basename", undef, "");
            if ($file)
            {
                my $tags=$file->get("Desktop Entry", "X-Created-By", "");
                foreach my $tag (split /;+/, $tags)
                {
                    $file->{_vfc_tags}->{$tag}=1;
                }
                my $categories=$file->get("Desktop Entry", "Categories", "");
                foreach my $category (split /;+/, $categories)
                {
                    $file->{_vfc_categories}->{$category}=1;
                }
            }
            else
            {
                cxlog("unable to read '$dir/$basename'\n");
            }
            return $file;
        }
    }
    return undef;
}

sub vfc_file_get_tags($$)
{
    my ($self, $file)=@_;

    return keys %{$file->{_vfc_tags}};
}

sub vfc_file_has_tag($$$)
{
    my ($self, $file, $tag)=@_;

    return $file->{_vfc_tags}->{$tag};
}

sub vfc_file_set_tags($$\@)
{
    my ($self, $file, $tags)=@_;

    my $modified;
    my $filename=$file->get_filename();
    foreach my $tag (keys %{$file->{_vfc_tags}})
    {
        if (!$tags->{$tag})
        {
            delete $self->{vfc_tags}->{$tag}->{$filename};
            delete $self->{_vfc_tags}->{$tag};
            $modified=1;
        }
    }
    foreach my $tag (keys %$tags)
    {
        if (!$file->{_vfc_tags}->{$tag})
        {
            $self->{vfc_tags}->{$tag}->{$filename}=$file;
            $file->{_vfc_tags}->{$tag}=1;
            $modified=1;
        }
    }
    if ($modified)
    {
        $file->set("Desktop Entry", "X-Created-By", join(";", sort keys %$tags));
    }
}

sub vfc_file_untag($$$)
{
    my ($self, $file, $pattern)=@_;

    my $still_has_tags;
    my $filename=$file->get_filename();
    foreach my $tag (keys %{$file->{_vfc_tags}})
    {
        if ($tag =~ /^$pattern/)
        {
            delete $self->{vfc_tags}->{$tag}->{$filename};
            delete $self->{_vfc_tags}->{$tag};
        }
        else
        {
            $still_has_tags=1;
        }
    }

    return $still_has_tags;
}

sub vfc_get_tags($)
{
    my ($self)=@_;

    $self->vfc_init() if (!defined $self->{vfc_files});
    return keys %{$self->{vfc_tags}};
}

sub vfc_get_tag_files($$)
{
    my ($self, $tag)=@_;

    $self->vfc_init() if (!defined $self->{vfc_files});
    return values %{$self->{vfc_tags}->{$tag}};
}

sub vfc_file_get_categories($$)
{
    my ($self, $file)=@_;

    return keys %{$file->{_vfc_categories}};
}

sub vfc_file_has_category($$$)
{
    my ($self, $file, $category)=@_;

    return $file->{_vfc_categories}->{$category};
}

sub vfc_get_categories($)
{
    my ($self)=@_;

    $self->vfc_init() if (!defined $self->{vfc_files});
    return keys %{$self->{vfc_categories}};
}

sub vfc_get_category_files($$)
{
    my ($self, $category)=@_;

    $self->vfc_init() if (!defined $self->{vfc_files});
    return values %{$self->{vfc_categories}->{$category}};
}

sub vfc_delete_file($$)
{
    my ($self, $file)=@_;

    my $filename=$file->get_filename();
    if (defined $self->{vfc_files})
    {
        my $basename=$filename;
        $basename=~ s%^.*/%%;
        delete $self->{vfc_files}->{$basename};
        foreach my $tag (keys %{$file->{_vfc_tags}})
        {
            delete $self->{vfc_tags}->{$tag}->{$filename};
        }
        foreach my $category (keys %{$file->{_vfc_categories}})
        {
            delete $self->{vfc_categories}->{$category}->{$filename};
        }
    }
    require CXRWConfig;
    CXRWConfig::uncache_file($filename);
    if (-f $filename)
    {
        cxlog("Deleting '$filename'\n");
        if (!unlink $filename)
        {
            cxerr("unable to delete '$filename': $!\n");
        }
    }
    return $file;
}


#####
#
# CXFolder
# This class is a wrapper around the XML Folder object.
# It keeps track of whether this is a CrossOver folder or not,
# and its category and tags. It also implements any change both in the
# XML file and in the relevant .directory file.
#
#####

{
    package CXFolder;
    use CXXMLDOM;
    use CXLog;

    sub new($$$$)
    {
        my ($class, $vfolder, $folder, $cxparent)=@_;
        my $self;

        $self->{vfolder}=$vfolder;
        $self->{xml}=$folder;
        if ($cxparent)
        {
            my $text=get_child($folder, "Name");
            $self->{mname}=CXXMLDOM::unicode_to_utf8(get_cdata($text));
            $self->{mpath}="$cxparent->{mpath}/$self->{mname}";
            $self->{mangled}="$cxparent->{mangled}_$self->{mname}";
        }
        else
        {
            # The top-level folder has a name we don't care about
            $self->{mname}="";
            $self->{mpath}="";
            $self->{mangled}="";
        }
        bless $self, $class;
        return $self;
    }

    sub init_category($$)
    {
        my ($self, $category)=@_;
        $self->{category}=$category;
        $self->{is_crossover}=1;
    }

    sub _parse_query()
    {
        my ($self)=@_;

        $self->{is_crossover}=0;
        $self->{query}=get_child($self->{xml}, "Query");
        return if (!defined $self->{query});

        my @extra;
        foreach my $node ($self->{query}->getElementsByTagName("Keyword"))
        {
            my $str=get_cdata($node);
            # Allow either '-' or '_' to take legacy menus into account
            if ($str =~ /^X-CXTree[-_]/)
            {
                if (defined $self->{category} and $self->{category} ne $str)
                {
                    cxlog("keyword mismatch: got '$self->{category}' then '$str'\n");
                }
                else
                {
                    $self->{category}=$str;
                    $self->{is_crossover}=1;
                }
            }
            elsif ($str =~ s/^X-//)
            {
                $self->{tags}->{$str}=1;
            }
            else
            {
                push @extra, $str;
            }
        }
        if ($self->{is_crossover} and @extra)
        {
            cxwarn("unexpected keywords found for '$self->{mpath}': @extra\n");
        }

    }

    sub get_xml($)
    {
        my ($self)=@_;
        return $self->{xml};
    }

    sub get_mname($)
    {
        my ($self)=@_;
        return $self->{mname};
    }

    sub get_mpath($)
    {
        my ($self)=@_;
        return $self->{mpath};
    }

    sub get_mangled($)
    {
        my ($self)=@_;
        return $self->{mangled};
    }

    sub get_category($)
    {
        my ($self)=@_;

        $self->_parse_query() if (!defined $self->{is_crossover});
        return $self->{category};
    }

    sub is_crossover($)
    {
        my ($self)=@_;

        $self->_parse_query() if (!defined $self->{is_crossover});
        return $self->{is_crossover};
    }

    sub get_tags($)
    {
        my ($self)=@_;

        $self->_parse_query() if (!defined $self->{is_crossover});
        return keys %{$self->{tags}};
    }

    sub has_tag($$)
    {
        my ($self, $tag)=@_;

        $self->_parse_query() if (!defined $self->{is_crossover});
        return $self->{tags}->{$tag};
    }

    sub add_tag($$)
    {
        my ($self, $tag)=@_;

        $self->_parse_query() if (!defined $self->{is_crossover});
        if (!$self->{tags}->{$tag})
        {
            $self->{tags}->{$tag}=1;
            $self->{modified}=1;
        }
    }

    sub remove_tag($$)
    {
        my ($self, $tag)=@_;

        $self->_parse_query() if (!defined $self->{is_crossover});
        if (exists $self->{tags}->{$tag})
        {
            delete $self->{tags}->{$tag};
            $self->{modified}=1;
        }
    }

    sub get_basename()
    {
        my ($self)=@_;
        my $desktop=get_child($self->{xml}, "Desktop");
        return ($desktop ? get_cdata($desktop) : "");
    }

    sub get_directory_file($)
    {
        my ($self)=@_;
        if (!$self->{directory})
        {
            my $basename=$self->get_basename() || "CXTree$self->{mangled}.directory";
            $self->{directory}=$self->{vfolder}->vfc_get_file($basename);
        }
        return $self->{directory};
    }

    sub set_directory_file($$)
    {
        my ($self, $directory)=@_;
        $self->{directory}=$directory;
    }

    sub save($)
    {
        my ($self)=@_;

        if ($self->{modified})
        {
            cxlog("Saving folder $self->{mpath}\n");

            # Rewrite the Folder's Query section
            my $query=get_child($self->{xml}, "Query");
            if (defined $query)
            {
                # We will rebuild this section from scratch
                while (1)
                {
                    my $n=$query->getFirstChild();
                    last if (!defined $n);
                    $query->removeChild($n);
                }
            }
            else
            {
                my $pos=get_child($self->{xml}, "Folder");
                $query=add_new_element($self->{xml}, "Query", undef, $pos);
            }

            my $or=add_new_element($query, "Or");
            add_new_element($or, "Keyword", $self->{category});
            my @tag_list=sort keys %{$self->{tags}};
            if (@tag_list != 0)
            {
                my $and=add_new_element($or, "And");
                add_new_element($and, "Keyword", $self->{category});
                foreach my $tag (@tag_list)
                {
                    add_new_element($and, "Keyword", "X-$tag");
                }
            }
        }

        my $file=$self->get_directory_file();
        if (!$file)
        {
            cxlog("unable to find the .directory file for '$self->{mpath}'\n");
        }
        elsif ($self->{modified} or $file->is_modified())
        {
            # Update the .directory file's X-Created-By field
            $self->{vfolder}->vfc_file_set_tags($file, $self->{tags});
            if (!$file->save())
            {
                cxerr("unable to write to '", $file->get_filename(), "': $!\n");
            }
        }
        $self->{modified}=0;
    }

    sub delete($)
    {
        my ($self)=@_;
        $self->_parse_query() if (!defined $self->{is_crossover});
        if (!$self->{is_crossover})
        {
            cxerr("refusing to delete non CrossOver folder '$self->{mpath}'\n");
            return;
        }

        # Delete the <Folder> tag
        cxlog("Deleting folder '$self->{mpath}'\n");
        # Note the impact on removeall_folder()'s iteration code
        remove_element($self->{xml}->getParentNode(), $self->{xml});

        # Delete the .directory file
        my $file=$self->get_directory_file();
        $self->{vfolder}->vfc_delete_file($file) if (defined $file);
    }
}


#####
#
# Folder helper functions
#
#####

sub get_folder_by_mpath($$)
{
    my ($self, $mpath)=@_;

    my $folder=$self->{folders}->{$mpath};
    # Check the parent node to verify that the folder has not been deleted
    return $folder if ($folder and $folder->getParentNode());

    $folder=get_tag_by_path($self->{menu_root}, "Folder", $mpath);
    $self->{folders}->{$mpath}=$folder;
    cxlog("353: found '$mpath'\n"); # FIXME: Debug
    return $folder;
}

sub get_cxfolder($$)
{
    my ($self, $folder)=@_;
    my $cxfolder=$self->{cxfolders}->{$folder};
    if (!$cxfolder)
    {
        my $cxparent;
        my $parent=$folder->getParentNode();
        if ($parent and $parent->getNodeName() eq "Folder")
        {
            $cxparent=$self->get_cxfolder($parent);
        }
        $cxfolder=CXFolder->new($self, $folder, $cxparent);
        $self->{cxfolders}->{$folder}=$cxfolder;
    }
    return $cxfolder;
}

sub create_cxfolder($$$)
{
    my ($self, $cxparent, $component)=@_;

    my $folder=$self->{folders}->{$component->{_mpath}};
    if (defined $folder and $component->{intermediate})
    {
        # We have already processed this folder so get out now
        return $self->get_cxfolder($folder);
    }
    cxlog("Creating folder '$component->{_mpath}'\n");

    # Update the XML tree
    my $cxfolder;
    $folder=get_child_by_name($cxparent->get_xml(), "Folder", $component->{_mname}) if (!$folder);
    if (defined $folder)
    {
        cxlog(" -> found existing folder\n");

        # Don't tag non-CrossOver menus
        $cxfolder=$self->get_cxfolder($folder);
        return $cxfolder if (!$cxfolder->is_crossover());
    }
    else
    {
        # Create the XML tag
        my $parent=$cxparent->get_xml();
        my $pos=get_child($parent, "Folder");
        $folder=add_new_element($parent, "Folder", undef, $pos);
        # VFolder is confused if it finds a '%' in the XML file and using the
        # XML character entity instead does not help. Furthermore the Name
        # field in the .directory file will override the XML <Name> tag
        # anyway, so let's just use the mangled folder name.
        add_new_element($folder, "Name", $component->{_mname});
        if ($self->uses_parent_tags())
        {
            set_element($folder, "Parent", "applications-all-users://$component->{_mpath}");
        }
        set_element($folder, "Desktop", "CXTree$component->{_mangled}.directory");
        $cxfolder=$self->get_cxfolder($folder);
        $cxfolder->init_category("X-CXTree$component->{_mangled}");
    }
    $cxfolder->add_tag($self->{tag});
    $self->{folders}->{$component->{_mpath}}=$folder;

    # (Re)Create the corresponding directory file (from scratch)
    # kg_create_directory() takes care of everything, we just need to figure
    # out the filename
    my $file=$cxfolder->get_directory_file();
    if (!defined $file)
    {
        my $dir=$self->get_desktop_dir();
        if (!defined $dir)
        {
            cxerr("unable to find the location of .directory files\n");
        }
        elsif (!-d $dir and !cxmkpath($dir))
        {
            cxerr("unable to create the '$dir' directory: $@\n");
        }
        else
        {
            my $filename="$dir/" . $cxfolder->get_basename();
            my $directory=CXMenu::kg_create_directory($filename, $component, 0);
            $cxfolder->set_directory_file($directory);
        }
    }
    elsif (!$component->{intermediate})
    {
        CXMenu::kg_create_directory($file->get_filename(), $component, 0);
    }
    $cxfolder->save();

    $self->{modified}=1;
    return $cxfolder;
}

sub removeall_folder($$$$)
{
    my ($self, $cxfolder, $pattern, $delete)=@_;

    cxlog("removeall_folder '", $cxfolder->get_mpath(), "'\n") if (CXLog::is_on());

    # Recurse in sub-folders
    # CXFolder::delete() deletes the folder and the previous indentation text
    # so the loop must go forward.
    my $in_use=0;
    my $node=$cxfolder->get_xml()->getFirstChild();
    while ($node)
    {
        # Get the next node before $node gets deleted
        my $next=$node->getNextSibling();
        if ($node->getNodeName() eq "Folder")
        {
            my $cxnode=$self->get_cxfolder($node);
            $in_use=1 if ($self->removeall_folder($cxnode, $pattern, $delete));
        }
        $node=$next;
    }
    cxlog("found subfolders\n") if ($in_use);

    if ($delete or !$in_use)
    {
        # Remove all that folder's .desktop files
        my $category=$cxfolder->get_category() ||
                     ("X-CXTree" . $cxfolder->get_mangled());
        foreach my $file ($self->vfc_get_category_files($category))
        {
            if (grep /^$pattern/, $self->vfc_file_get_tags($file))
            {
                $self->vfc_delete_file($file);
            }
            else
            {
                $in_use=1;
                cxlog("in-use (due to ", $file->get_filename(), ")\n") if (CXLog::is_on());
                last if (!$delete);
            }
        }
        cxlog("found menus\n") if ($in_use);
    }

    if (!$cxfolder->is_crossover())
    {
        # This folder does not belong to us so don't touch it
        return 1;
    }

    # Delete or untag the folder
    my $had_tags;
    my $still_has_tags;
    foreach my $tag ($cxfolder->get_tags())
    {
        cxlog("tag=[$tag]\n"); # FIXME: Debug
        if ($tag =~ /^$pattern/)
        {
            $cxfolder->remove_tag($tag);
            $had_tags=1;
        }
        else
        {
            $still_has_tags=1;
        }
    }
    if (!$still_has_tags and !$in_use)
    {
        # Delete the folder
        delete $self->{folders}->{$cxfolder->get_mpath()};
        $cxfolder->delete();
        $self->{modified}=1;
    }
    elsif ($had_tags)
    {
        $cxfolder->save();
        $self->{modified}=1;
        $in_use=1;
    }

    return $in_use;
}


#####
#
# Main VFolder XML file creation and parsing
#
#####

my $broken_user_menus;
sub has_broken_user_menus($)
{
    my ($self)=@_;

    if (!defined $broken_user_menus)
    {
        # Notes:
        # - In early Gnome 2 versions (up to 2.2.0), users cannot customize
        #   the menus! They have to make a complete copy of the system-wide
        #   menu structure.
        # - On Red Hat, even with version 2.4.2 the user must make a full
        #   copy of the system-wide file!
        # - But if a user installs the Ximian Gnome packages on Red Hat then
        #   this problem is fixed if the version is greater than 2.2.0.
        #   However the output of 'gnome-panel --version' does not make it
        #   possible to distinguish between the Red Hat and Ximian packages.
        # - Finally, we cannot rely on 'rpm -q' because this does not work
        #   when invoked from our RPM's postinstall script, and because it
        #   does not work on non-RPM based systems.

        my $version;
        if (-f "/etc/redhat-release" and !-f "/etc/mandrake-release")
        {
            # On Red Hat try 'rpm -q' first as it is the most accurate
            my $str=cxbackquote("rpm -q gnome-panel 2>/dev/null");
            $version=$1 if ($str =~ /^gnome-panel-/);
        }
        if (!defined $version)
        {
            # Otherwise, use locate_gui.sh's Gnome version first
            $version=$self->{gnome_version};
            if (!defined $version)
            {
                # And as the last chance, try 'rpm -q'
                my $str=cxbackquote("rpm -q gnome-panel 2>/dev/null");
                $version=$1 if ($str =~ /^gnome-panel-/);
            }
        }
        $version||=""; # assume menus are not broken
        if ($version =~ /^2\.(?:[01]\.|2\.0)/ or
            (-f "/etc/redhat-release" and !-f "/etc/mandrake-release" and
             $version !~ /ximian/))
        {
            $broken_user_menus=1;
        }
        else
        {
            $broken_user_menus=0;
        }
    }
    return $broken_user_menus;
}

sub init_xml($$)
{
    my ($self)=@_;

    return $self->{xml} if (defined $self->{xml});
    $self->{xml}=0;

    my $xml;
    if (-f $self->{menu})
    {
        $xml=CXXMLDOM::parse_xml_file($self->{menu});
    }
    elsif ($self->has_broken_user_menus())
    {
        cxlog("Copying the system VFolder file\n");
        $xml=CXXMLDOM::parse_xml_file($self->{managed_menu});

        if (defined $xml)
        {
            # Tag each Folder with a comment saying it comes from the
            # system file. Don't set 'modified' here so XML file gets
            # saved (i.e. copied) only if we do other changes.
            foreach my $folder ($xml->getElementsByTagName("Folder"))
            {
                my $comment=XML::DOM::Comment->new($xml);
                $comment->setData("X-crossover-managed");
                $folder->insertBefore($comment, $folder->getFirstChild());
            }
        }
    }
    else
    {
        # This is done based off of the file that seems to be automatically
        # created by Gnome 2.2.2.  It should be possible to get Gnome
        # to generate it for us, but as of 7/12/03, JPW didn't know how.
        cxlog("Creating per-user VFolder file\n");
        my $dir=cxdirname($self->{menu});
        my $vfolder="<VFolderInfo>\n" .
                    "  <WriteDir>$dir</WriteDir>\n" .
                    "  <ItemDir>$dir</ItemDir>\n" .
                    "  <Folder>\n" .
                    "    <Name>Root</Name>\n" .
                    "    <Parent>applications-all-users:///</Parent>\n" .
                    "  </Folder>\n" .
                    "</VFolderInfo>\n";
        $xml=CXXMLDOM::parse_xml_string($vfolder);
    }
    return 0 if (!$xml);

    # Now check that the basic structure of this XML file is correct
    my $vfolderinfo=find_tag($xml, "VFolderInfo");
    if (!$vfolderinfo)
    {
        cxerr("unable to find the VFolderInfo tag in '$self->{menu}'\n");
        return 0;
    }

    my $menu_root;
    my $node=$vfolderinfo->getFirstChild();
    while (defined $node)
    {
        my $name=$node->getNodeName();
        if ($name eq "DesktopDir")
        {
            cxlog("DesktopDir='", get_cdata($node), "'\n");
            push @{$self->{desktop_dirs}}, get_cdata($node);
        }
        elsif ($name eq "WriteDir")
        {
            cxlog("WriteDir='", get_cdata($node), "'\n");
            push @{$self->{write_dirs}}, get_cdata($node);
        }
        elsif ($name =~ /^(?:Item|Merge)Dir$/)
        {
            cxlog("OtherDir='", get_cdata($node), "'\n");
            push @{$self->{other_dirs}}, get_cdata($node);
        }
        elsif ($name eq "Folder" and !defined $menu_root)
        {
            my $text=get_child($node, "Name");
            $menu_root=$node if (get_cdata($text) =~ /^(Applications|Root|SuSE)$/);
        }
        $node=$node->getNextSibling();
    }
    if (!$menu_root)
    {
        cxerr("unable to find the menu root in '$self->{menu}'\n");
        return 0;
    }

    $self->{xml}=$xml;
    $self->{menu_root}=$menu_root;
    $self->{vfolderinfo}=$vfolderinfo;

    return 1;
}


#####
#
# Main
#
#####


sub detect($$$$)
{
    my ($class, $cxoptions, $cxconfig, $gui_info)=@_;
    return () if (!$gui_info->{vfolder_on});

    # Check for XML::DOM support
    my $err=CXXMLDOM::get_xml_load_error();
    if ($err)
    {
        cxerr("unable to use XML::DOM:\n$err\n");
        return ();
    }

    # Note: CXMenuVFolder does not handle the VFolder desktop icons as these
    # have nothing in common with the regular VFolder menus. Instead this
    # task is delegated to CXMenuVFDesktop.
    my @selves;
    if ($gui_info->{vfolder_preferred_menu})
    {
        my $self={
            tag              => $cxoptions->{tag},
            menu             => $gui_info->{vfolder_preferred_menu},
            managed_menu     => $gui_info->{vfolder_global_menu},
            gnome_version    => $gui_info->{gnome_version}
        };
        bless $self, $class;
        push @selves, $self;
    }
    if ($gui_info->{vfolder_preferred_alt_menu})
    {
        my $self={
            tag              => $cxoptions->{tag},
            menu             => $gui_info->{vfolder_preferred_alt_menu},
            managed_menu     => $gui_info->{vfolder_global_menu},
            gnome_version    => $gui_info->{gnome_version}
        };
        bless $self, $class;
        push @selves, $self;
    }
    return @selves;
}

sub id($)
{
    my ($self)=@_;
    my $id="CXMenuVFolder/$self->{menu}";
    $id =~ s%/+%/%g;
    return $id;
}

sub install($$)
{
    my ($self, $components)=@_;

    my $menu=@$components[-1];
    return 1 if ($menu->{is_desktop});
    return 0 if (!$self->init_xml());

    my $mpath="";
    my $mangled="";
    my $cxfolder=$self->get_cxfolder($self->{menu_root});
    foreach my $component (@$components)
    {
        $component->{_mname}=mangle_string($component->{name});
        $mpath="$mpath/$component->{_mname}";
        $component->{_mpath}=$mpath;
        $mangled="${mangled}_$component->{_mname}";
        $component->{_mangled}=$mangled;
        $component->{_encoding}="UTF-8";
        $component->{_creator}="";
        last if (!$component->{is_dir});
        $cxfolder=$self->create_cxfolder($cxfolder, $component);
        return 0 if (!defined $cxfolder);
    }

    if ($menu->{is_dir})
    {
        # Make sure our brand new (maybe empty) folder will not be
        # 'garbage collected' by finalize()
        $self->{status}->{$menu->{_mpath}}="in-use";
    }
    else
    {
        my $mdir=$menu->{_mpath};
        $mdir=~s%/[^/]*$%%;
        $self->{status}->{$mdir}="in-use";
        # We must check that the parent folder belongs to us. This is
        # because we cannot create menus in folders that don't belong to us
        # since we don't know what combination of categories would be needed
        # to satisfy the <query> condition.
        if (!$cxfolder->is_crossover())
        {
            cxerr("unable to create '$menu->{path}' because it is not in a CrossOver folder\n");
            return 0;
        }

        # Create the desktop file
        my $dir=$self->get_write_dir();
        if (!defined $dir)
        {
            cxerr("unable to find the location of .desktop files\n");
            return 0;
        }
        my $filename="$dir/CXTree$menu->{_mangled}.desktop";
        # We don't need _escaping (at least for Gnome)
        $menu->{_section_name}="Desktop Entry";
        $menu->{_categories}="Application;" . $cxfolder->get_category() . ";";
        $menu->{_creator}=" (" . $self->id() . ")" if ($ENV{CX_TAGALL});
        my $file=CXMenu::kg_create_desktop_file($filename, $menu);
        $self->vfc_add_file($file);
    }

    return 1;
}

sub query($$)
{
    my ($self, $components)=@_;
    return ($self->id(), "") if (!defined $components);

    my $menu=@$components[-1];
    return "" if ($menu->{is_desktop});
    return "0" if (!$self->init_xml());

    # Check that everything is in place in the XML file
    my $mpath=$menu->{is_dir} ? $menu->{path} : $menu->{dir};
    $mpath=mangle_path($mpath);
    my $folder=$self->get_folder_by_mpath($mpath);
    return "" if (!$folder);

    # Check that the folder belongs to us
    my $cxfolder=$self->get_cxfolder($folder);
    return "" if (!$cxfolder->is_crossover());

    # Then check the presence of the .desktop or .directory file
    if ($menu->{is_dir})
    {
        return "" if (!grep /^$self->{tag}$/, $cxfolder->get_tags());
        my $file=$cxfolder->get_directory_file();
        return "" if (!defined $file);
        cxlog("Checking for '", $file->get_filename(), "'\n") if (CXLog::is_on());

        # Check the X-Created-By field
        if ($self->vfc_file_has_tag($file, $self->{tag}))
        {
            return $self->id();
        }
    }
    else
    {
        my $mangled=mangle_string($menu->{path});
        my $basename="CXTree$mangled.desktop";
        my $file=$self->vfc_get_file($basename);
        return "" if (!defined $file);
        cxlog("Checking for '", $file->get_filename(), "'\n") if (CXLog::is_on());

        # Check the X-Created-By field
        if (!$self->vfc_file_has_tag($file, $self->{tag}))
        {
            return "";
        }

        # Also check the Categories field
        if ($self->vfc_file_has_category($file, $cxfolder->get_category()))
        {
            return $self->id();
        }
    }

    return "";
}

sub uninstall($$)
{
    my ($self, $components)=@_;

    my $menu=@$components[-1];
    return 1 if ($menu->{is_desktop});
    return 0 if (!$self->init_xml());

    # Always mark the parent folder for garbage collection
    $self->{garbage_collect}->{$menu->{dir}}=1;

    if ($menu->{is_dir})
    {
        my $mpath=mangle_path($menu->{path});
        my $folder=$self->get_folder_by_mpath($mpath);
        return 1 if (!$folder);
        my $cxfolder=$self->get_cxfolder($folder);
        $self->removeall_folder($cxfolder, "$self->{tag}\$", 1);
    }
    else
    {
        my $mangled=mangle_string($menu->{path});
        my $basename="CXTree$mangled.desktop";
        my $file=$self->vfc_get_file($basename);
        if ($file and grep /^$self->{tag}$/, $self->vfc_file_get_tags($file))
        {
            # When uninstalling menus, it's enough that they belong to the
            # current bottle. There's no need to cross-check the Categories
            # field.
            $self->vfc_delete_file($file);
        }
    }

    return 1;
}

sub removeall($$)
{
    my ($self, $pattern)=@_;
    return 0 if (!$self->init_xml());
    $pattern=CXUtils::get_product_id() . "\$" if ($pattern eq "legacy");

    my $cxfolder=$self->get_cxfolder($self->{menu_root});
    $self->removeall_folder($cxfolder, $pattern, 0);

    # Go through these after removeall_folder() has done its thing (and
    # potentially re-synced the .directory file's tags from XML).
    foreach my $tag ($self->vfc_get_tags())
    {
        cxlog("removeall(): tag=[$tag]\n");
        next if ($tag !~ /^$pattern/);
        foreach my $file ($self->vfc_get_tag_files($tag))
        {
            my $filename=$file->get_filename();
            if ($filename =~ /\.desktop$/)
            {
                $self->vfc_delete_file($file);
            }
            elsif (!$self->vfc_file_untag($file, $pattern))
            {
                # The .directory file is not in use anymore
                $self->vfc_delete_file($file);
            }
            elsif (!$file->save())
            {
                cxerr("unable to save '$filename': $!\n");
            }
        }
    }

    return 1;
}

sub gc_get_cxfolder_status($$)
{
    my ($self, $cxfolder)=@_;

    # No need to look any further if the folder does not belong to this bottle
    if (!$cxfolder->is_crossover() or !$cxfolder->has_tag($self->{tag}))
    {
        return "alien";
    }

    # Check the folder's .desktop files
    my $status="empty";
    foreach my $file ($self->vfc_get_category_files($cxfolder->get_category()))
    {
        if ($self->vfc_file_has_tag($file, $self->{tag}))
        {
            cxlog("in-use (due to ", $file->get_filename(), ")\n") if (CXLog::is_on());
            return "in-use";
        }
        cxlog("non-empty (due to ", $file->get_filename(), ")\n") if (CXLog::is_on());
        $status="non-empty";
    }

    # Check the sub-folders
    my $child=$cxfolder->get_xml()->getFirstChild();
    while ($child)
    {
        if ($child->getNodeName() eq "Folder")
        {
            my $cxchild=$self->get_cxfolder($child);
            my $mpath=$cxchild->get_mpath();
            my $rc=$self->{status}->{$mpath} ||
                   $self->gc_get_cxfolder_status($cxchild);
            if ($rc eq "in-use")
            {
                cxlog("in-use (due to folder $mpath)\n");
                return "in-use";
            }
            cxlog("non-empty (due to folder $mpath)\n");
            $status="non-empty";
        }
        $child=$child->getNextSibling();
    }

    return $status;
}

sub gc_get_folder_status($$)
{
    my ($self, $path)=@_;

    my $mpath=mangle_path($path);
    my $folder=$self->get_folder_by_mpath($mpath);
    if (!$folder)
    {
        cxlog("folder '$path' has already been deleted\n");
        return undef;
    }
    my $cxfolder=$self->get_cxfolder($folder);
    my $status=$self->gc_get_cxfolder_status($cxfolder);
    return ($status, $cxfolder);
}

sub gc_delete_folder($$$)
{
    my ($self, $path, $cxfolder)=@_;

    delete $self->{folders}->{$cxfolder->get_mpath()};
    $cxfolder->delete();
    $self->{modified}=1;
}

sub gc_untag_folder($$$)
{
    my ($self, $path, $cxfolder)=@_;

    $cxfolder->remove_tag($self->{tag});
    $cxfolder->save();
    $self->{modified}=1;
}

sub finalize($)
{
    my ($self)=@_;
    return 0 if (!$self->{xml});

    my $status=$self->{status};
    my $may_be_empty=!($status and %$status);

    $self->SUPER::finalize();

    # Check the managed tags so we can delete the xml file if all that
    # remains are managed folders (see init_xml())
    my $all_managed;
    if ($self->{menu} ne $self->{managed_menu} and $may_be_empty and
        -f $self->{menu})
    {
        $all_managed=1;
        my @folders=($self->{menu_root});
        while (@folders)
        {
            my $folder=shift @folders;
            my $has_comment;
            my $node=$folder->getFirstChild();
            if (CXLog::is_on())
            {
                my $text=get_child($folder, "Name");
                cxlog("managed? ", get_cdata($text), "\n");
            }
            while ($node)
            {
                if ($node->getNodeName() eq "#comment" and
                    $node->getData() =~ /X-crossover-managed/i)
                {
                    $has_comment=1;
                }
                elsif ($node->getNodeName() eq "Folder")
                {
                    push @folders, $node;
                }
                $node=$node->getNextSibling();
            }
            if (!$has_comment)
            {
                cxlog(" -> no!\n");
                $all_managed=0;
                last;
            }
        }
    }

    if ($all_managed)
    {
        cxlog("Deleting '$self->{menu}'\n");
        if (!unlink $self->{menu})
        {
            cxerr("unable to delete '$self->{menu}': $!\n");
            return 0;
        }
    }
    elsif ($self->{modified})
    {
        return CXXMLDOM::save_xml_file($self->{menu}, $self->{xml},
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    }
    else
    {
        cxlog("XML file '$self->{menu}' not modified\n");
    }
    return 1;
}

return 1;
