# (c) Copyright 2009. CodeWeavers, Inc.

import os
import sys
import traceback
import urllib

import gobject
import gtk
import gtk.glade

import cxlog
import cxutils

# for localization
from cxutils import cxgettext as _



def toplevel_quit():
    """If there are no visible toplevel windows left, call gtk.main_quit().

    This should be called when a toplevel is closed on a process that might have
    multiple toplevels and needs to live as long as at least one is still open."""
    for toplevel in gtk.window_list_toplevels():
        if toplevel.get_property('visible'):
            break
    else:
        gtk.main_quit()



def get_crossover_icon():
    global crossover_icon
    try:
        return crossover_icon
    except NameError:
        icon_filename = os.path.join(cxutils.CX_ROOT, 'share', 'icons', 'crossover.xpm')
        try:
            crossover_icon = gtk.gdk.pixbuf_new_from_file(icon_filename)
        except gobject.GError:
            cxlog.warn("couldn't load icon file %s:\n%s" % (cxlog.debug_str(icon_filename), traceback.format_exc()))
            crossover_icon = None
        return crossover_icon



def set_default_icon():
    icon = get_crossover_icon()
    if icon:
        gtk.window_set_default_icon(icon)


def draw_widget_background(_widget, event, pixbuf):
    event.window.draw_pixbuf(None, pixbuf, 0, 0, 0, 0)
    return False


def set_widget_background(widget, filename):
    filename = os.path.join(cxutils.CX_ROOT, 'share', 'images', filename)
    try:
        pixbuf = gtk.gdk.pixbuf_new_from_file(filename)
    except gobject.GError:
        cxlog.warn("couldn't load icon file %s:\n%s" % (cxlog.debug_str(filename), traceback.format_exc()))
        return

    widget.connect('expose-event', draw_widget_background, pixbuf)



#####
#
# CXMessageDlg
#
#####

def _messagedlg_handle_response(dialog, response, response_function, close_on_response, user_data):
    if close_on_response:
        dialog.destroy()
    if response_function:
        if user_data is not None:
            result = response_function(response, user_data)
        else:
            result = response_function(response)
        if not close_on_response and result:
            dialog.destroy()

def CXMessageDlg(message=None, primary=None, secondary=None, markup=None, buttons=gtk.BUTTONS_OK, response_function=None, close_on_response=True, parent=None, user_data=None, button_array=None, message_type=None):
    if not markup:
        if message is None:
            text = []
            if primary is not None:
                text.append(u'<span weight="bold" size="larger">%s</span>' % cxutils.html_escape(primary))
            if secondary is not None:
                text.append(cxutils.html_escape(secondary))
            markup = '\n\n'.join(text)
        else:
            markup = cxutils.html_escape(message)
    if message_type is None:
        cxlog.warn("Callers to cxguitools.CXMessageDlg should always set an alert type.")
        message_type = gtk.MESSAGE_INFO
    if button_array:
        dialog = gtk.MessageDialog(buttons=gtk.BUTTONS_NONE, flags = gtk.DIALOG_MODAL, parent=parent, type=message_type)
        for buttonPair in button_array:
            dialog.add_button(buttonPair[0], buttonPair[1])
    else:
        dialog = gtk.MessageDialog(buttons=buttons, flags = gtk.DIALOG_MODAL, parent=parent, type=message_type)
    dialog.set_markup(markup)
    dialog.connect('response', _messagedlg_handle_response, response_function, close_on_response, user_data)
    dialog.show()


def normalizeBackgroundColor(widget):
    parent = widget.get_parent()
    widget.modify_base(gtk.STATE_NORMAL, parent.get_style().bg[gtk.STATE_NORMAL])


def _malware_response(response, user_data):
    c4pfile, stop_loop = user_data
    if response == 0:
        # More Info
        url = 'http://www.codeweavers.com/support/wiki/malware/' + urllib.quote_plus(c4pfile.malware_appid)
        cxutils.launch_url(url)
        return False
    else:
        # Close
        if stop_loop:
            gtk.main_quit()
        return True

def show_malware_dialog(c4pfile, parent=None, stop_loop=False):
    primary = _("The file '%(filename)s' contains malware") % {'filename': c4pfile.filename}
    secondary = _("This file cannot be used because it would install malicious software or otherwise harm your computer.")
    CXMessageDlg(secondary=secondary, primary=primary,
                                button_array=[[_(u"More Info\u2026"), 0],
                                              [gtk.STOCK_CLOSE, 1]],
                                user_data = (c4pfile, stop_loop),
                                response_function=_malware_response,
                                close_on_response=False,
                                parent=parent,
                                message_type = gtk.MESSAGE_ERROR)


#####
#
# CustomWidgets
#
#####

class CustomWidgets(object):
    @staticmethod
    def launch_url(_view, url, _type):
        cxutils.launch_url(url)


    @staticmethod
    def create_htmltextview(_str1, _str2, int1, _int2):
        import htmltextview
        widget = htmltextview.HtmlTextView()
        widget.show()

        widget.connect("url-clicked", CustomWidgets.launch_url)
        if int1:
            # Raising int1 specifies a label-like appearance
            #  and behavior for the widget.
            widget.connect("realize", normalizeBackgroundColor)
            widget.set_wrap_mode(gtk.WRAP_WORD)
            widget.set_editable(False)
            widget.set_cursor_visible(False)

        return widget


    @staticmethod
    def create_assoceditor(_str1, _str2, _int1, _int2):
        # Don't use a normal import statement to avoid a cyclic import complaint
        # from pylint. This is OK because we do it lazily.
        return __import__('assocwidget').AssocEditor()


    @staticmethod
    def create_menueditor(_str1, _str2, _int1, _int2):
        # Don't use a normal import statement to avoid a cyclic import complaint
        # from pylint. This is OK because we do it lazily.
        return __import__('menuwidget').MenuEditor()


    @staticmethod
    def create_plugineditor(_str1, _str2, _int1, _int2):
        # Don't use a normal import statement to avoid a cyclic import complaint
        # from pylint. This is OK because we do it lazily.
        return __import__('pluginwidget').PluginEditor()


    @staticmethod
    def custom_widget_handler(_glade, function_name, _widget_name, str1, str2, int1, int2):

        try:
            handler = getattr(CustomWidgets, function_name)
        except:
            traceback.print_exc()
            return None

        return handler(str1, str2, int1, int2)

#####
#
# Standard warning dialog if running as root
#
#####

def return_from_root_check(response):
    if response:
        sys.exit()

# This should be called after the main dialog (e.g. InstallerAssistantController)
#  has loaded, but before gtk.main() is invoked.
def warn_if_root():
    if os.getuid() == 0:
        # We probably shouldn't be running as root.
        warningString = _("Running this tool as root is very dangerous, and will only allow you to configure CrossOver for the root user.\n\nIf you wish to create bottles for system-wide access, log in as a user and make use of the 'Publish Bottle' feature.")
        CXMessageDlg(warningString, buttons=None, button_array=[[_("Exit"), 1], [_("Configure CrossOver for Root"), 0]], response_function=return_from_root_check, message_type=gtk.MESSAGE_WARNING)


#####
#
# Standard File Picker selectors
#
#####

FILTERS_RUNNABLE = set(('exe', 'lnk'))
FILTERS_INSTALLABLE = set(('exe', 'fonts', 'msi', 'zip'))
FILTERS_CXARCHIVES = set(('archives', 'cxarchives'))
FILTERS_ALLFILES = set(('all',))

def add_ipattern(file_picker, pattern):
    """Adds a case-insensitive pattern based on the given case-sensitive one.
    """
    insensitive = []
    for c in pattern:
        if c.isalpha():
            insensitive.append("[%s%s]" % (c.lower(), c.upper()))
        else:
            insensitive.append(c)
    file_picker.add_pattern("".join(insensitive))

def add_filters(file_picker, filters):
    """Adds the specified set of standard file filters to the file picker."""
    file_filters = {}
    default = ''

    if 'archives' in  filters:
        file_filter = gtk.FileFilter()
        # These are Unix files so keep them case-sensitive
        file_filter.add_pattern("*.cpio.7z")
        file_filter.add_pattern("*.cpio.gz")
        file_filter.add_pattern("*.cpio.bz2")
        file_filter.add_pattern("*.cpio.Z")
        file_filter.add_pattern("*.tar.7z")
        file_filter.add_pattern("*.tar.gz")
        file_filter.add_pattern("*.tar.bz2")
        file_filter.add_pattern("*.tar.Z")
        file_filter.add_pattern("*.tgz")
        file_filters[_("Unix Archives")] = file_filter

    if 'cxarchives' in  filters:
        file_filter = gtk.FileFilter()
        # These are Unix files so keep them case-sensitive
        file_filter.add_pattern("*.cxarchive")
        default = _("CrossOver Bottle Archives")
        file_filters[default] = file_filter

    if 'exe' in filters:
        file_filter = gtk.FileFilter()
        add_ipattern(file_filter, "*.bat")
        add_ipattern(file_filter, "*.cmd")
        add_ipattern(file_filter, "*.com")
        add_ipattern(file_filter, "*.exe")
        add_ipattern(file_filter, "autorun.inf")
        default = _("Windows Programs")
        file_filters[default] = file_filter

    if 'fonts' in filters:
        file_filter = gtk.FileFilter()
        add_ipattern(file_filter, "*.otf")
        add_ipattern(file_filter, "*.ttc")
        add_ipattern(file_filter, "*.ttf")
        file_filters[_("Windows Fonts")] = file_filter

    if 'lnk' in filters:
        file_filter = gtk.FileFilter()
        add_ipattern(file_filter, "*.lnk")
        file_filters[_("Windows Shortcuts")] = file_filter

    if 'msi' in filters:
        file_filter = gtk.FileFilter()
        add_ipattern(file_filter, "*.msi")
        add_ipattern(file_filter, "*.msp")
        file_filters[_("Windows Installer Packages")] = file_filter

    if 'zip' in filters:
        file_filter = gtk.FileFilter()
        add_ipattern(file_filter, "*.cab")
        add_ipattern(file_filter, "*.zip")
        add_ipattern(file_filter, "*.rar")
        file_filters[_("Windows Archives")] = file_filter

    if 'all' in filters:
        file_filter = gtk.FileFilter()
        add_ipattern(file_filter, "*.*")
        file_filters[_("All Files")] = file_filter

    names = file_filters.keys()
    names.sort()
    for name in names:
        file_filter = file_filters[name]
        file_filter.set_name(name)
        file_picker.add_filter(file_filter)
        if name == default:
            # Make this filter the default
            file_picker.set_filter(file_filter)

def show_help(page):
    locale_doc_dir = _("doc/en")
    url = os.path.join(cxutils.CX_ROOT, locale_doc_dir, "html", page)
    cxutils.launch_url(url)
