# (c) Copyright 2009-2010. CodeWeavers, Inc.

import errno
import fcntl
import locale
import os
import re
import stat
import subprocess
import time
import traceback

import cxlog
import distversion
import cxobjc

import gettext

class CXUtils(cxobjc.Proxy):
    pass

#####
#
# CX_ROOT
#
#####

# Points to CrossOver's root directory.
# This should be set at startup and is then considered to be constant for the
# duration of the process.
CX_ROOT = None



#####
#
# File and directory helpers
#
#####

def mkdirs(path, *args, **kwargs):
    """Similar to os.makedirs() but returns False to signal errors and logs
    errors.
    """
    try:
        os.makedirs(path, *args, **kwargs)
    except OSError, ose:
        if ose.errno != errno.EEXIST:
            cxlog.err("unable to create the '%s' directory: %s" % (cxlog.debug_str(path), ose.strerror))
            return False
    return True


#####
#
# Cross-process lock
#
#####

class FLock(object):

    def __init__(self, name):
        """Creates and takes a lock called name using flock().

        This is used to avoid races between processes.
        """
        self._name = name

        # Create the lock directory and perform some sanity checks
        lockdir = os.path.join(os.environ.get('TMPDIR', '/tmp'),
                               ".wine-%d" % os.geteuid())
        mkdirs(lockdir, 0700)
        st = os.lstat(lockdir)
        if not stat.S_ISDIR(st.st_mode):
            raise ValueError("'%s' is not owned by you" % lockdir)
        if st.st_uid != os.geteuid():
            raise ValueError("'%s' is not owned by you" % lockdir)
        if st.st_mode & 077:
            raise ValueError("'%s' must not be accessible by other users" % lockdir)

        # Perform some more sanity checks and create the lock file
        self._path = os.path.join(lockdir, name + ".lock")
        try:
            st = os.lstat(self._path)
            if not stat.S_ISREG(st.st_mode):
                raise ValueError("'%s' is not a regular file" % self._path)
        except OSError, ose:
            if ose.errno != errno.ENOENT:
                raise ose
        self._lock = open(self._path, 'w')

        # Take the lock
        cxlog.log("%d: Grabbing the '%s' lock" % (os.getpid(), self._path))
        try:
            fcntl.flock(self._lock, fcntl.LOCK_EX)
        except Exception, e:
            # The filesystem probably does not support locking
            cxlog.warn("unable to take the '%s' lock: %s" % (name, e))
            self._lock.close()
            self._lock = None
            raise # re-raise the exception

    def _getname(self):
        """Returns the lock name."""
        return self._name

    name = property(_getname)

    def locked(self):
        """Returns True if the lock has not been released yet."""
        return self._lock is not None

    def release(self):
        """Releases the lock.

        Note that deleting a lock carries a very high risk of causing races
        outside the scope of this class and has essentially no benefit. So the
        file created for the lock is not deleted and no support is provided
        for deleting it.
        """
        if self._lock:
            cxlog.log("%d: Releasing the '%s' lock" % (os.getpid(), self._name))
            try:
                try:
                    fcntl.flock(self._lock, fcntl.LOCK_UN)
                except Exception, e:
                    cxlog.warn("unable to release the '%s' lock: %s" % (self._name, e))
            finally:
                self._lock.close()
                self._lock = None


#####
#
# String handling
#
#####

def unescape_string(string):
    """Unescapes a string that was placed in double-quotes."""
    string = string.replace('\\\"', '\"')
    string = string.replace("\\\\", "\\")
    return string


def escape_string(string):
    """Escapes the string so it can be placed in double-quotes."""
    string = string.replace("\\", "\\\\")
    string = string.replace('"', '\\"')
    return string


def shquote_string(string):
    """Quote a string so it can be used in shell commands.

    Note that this implies escaping '$'s and '`'s which may not be appropriate
    in another context.
    """
    string = string.replace('\\', '\\\\')
    string = string.replace('\$', '\\\$')
    string = string.replace('\"', '\\\"')
    string = string.replace("\'", "\\\'")
    return '"' + string + '"'

_BAD_SHELL_CHARS = re.compile("[^a-zA-Z0-9\/.,+_-]")
def maybe_shquote_string(string):
    if _BAD_SHELL_CHARS.search(string):
        return shquote_string(string)
    return string

def mangle(string, characters):
    result = []
    for char in string:
        if char in characters or not (0x1f < ord(char) < 0x80):
            if char == ' ':
                result.append('+')
            elif char == '/':
                result.append('_')
            elif char == '\\':
                result.append('~')
            else:
                result.append('^%02X' % ord(char))
        else:
            result.append(char)
    return ''.join(result)

def mangle_path(string):
    return mangle(string, '\\ $"\'`<>#&:;%^_~+')

def unmangle(string):
    result = []
    x = 0
    while x < len(string):
        char = string[x]
        x += 1
        if char == '+':
            result.append(' ')
        elif char == '_':
            result.append('/')
        elif char == '~':
            result.append('\\')
        elif char == '^':
            try:
                result.append(chr(int(string[x:x+2], 16)))
            except ValueError:
                # not followed by valid hex digits
                pass
            else:
                x += 2
        else:
            result.append(char)
    return ''.join(result)


_ACCELERATOR_RE = re.compile('&(.)')

def remove_accelerators(string):
    """Remove the ampersands that identify the accelerator keys on Windows.

    While only the first one is meant to be significant, they should all be
    removed, and '&&' should be turned into a regular '&'.
    """
    return _ACCELERATOR_RE.sub('\\1', string)


def md5_hasher():
    """Returns an object for computing MD5 hashes."""
    try:
        # pylint: disable-msg=F0401,E1101
        import hashlib
        return hashlib.md5()
    except ImportError:
        import md5
        return md5.new()


def bottlename_character_filter(character):
    """Returns False for characters that are best avoided
    in bottle names.
    """
    # \/*?:|<>" are not valid characters in Windows paths
    # {} are not valid in paths stored in Debian and Solaris packages
    return character not in "\\/*?:|<>\"'{}"


@cxobjc.method(CXUtils, 'sanitizedBottleName_')
def bottlename_sanitize(bottlename):
    """Removes characters that are best avoided
      in bottle names. """

    result = []
    for x in bottlename:
        if bottlename_character_filter(x):
            result.append(x)
        else:
            result.append('_')

    # . causes the bottle to be hidden if it's the first character
    return ''.join(result).lstrip(".")

def html_escape(string):
    """Escape characters that have special meaning in html."""
    replacements = {
        '<': '&lt;',
        '>': '&gt;',
        '&': '&amp;'}
    result = []
    for char in string:
        result.append(replacements.get(char, char))
    return ''.join(result)

@cxobjc.method(CXUtils, 'cmdLineToArgv_')
def cmdlinetoargv(string):
    """Convert a windows-style quoted command line to an argument list. This is
    NOT equivalent to CommandLineToArgvW."""
    in_quotes = False
    pos = 0
    result = []
    arg = []
    while pos < len(string):
        char = string[pos]
        if char == ' ' and not in_quotes:
            if arg:
                result.append(''.join(arg))
                del arg[:]
        elif char == '\\' and pos+1 < len(string) and string[pos+1] in '\\"':
            pos += 1
            arg.append(string[pos])
        elif char == '"':
            in_quotes = not in_quotes
        else:
            arg.append(string[pos])
        pos += 1
    if arg:
        result.append(''.join(arg))
    return result

def argvtocmdline(argv):
    """Convert an argument list to a windows-style quoted command line."""
    result = []
    for arg in argv:
        arg = arg.replace("\\", "\\\\").replace('"', '\\"')
        if ' ' in arg:
            arg = '"%s"' % arg
        result.append(arg)
    return ' '.join(result)

@cxobjc.method(CXUtils, 'basename_')
def basename(path):
    """Get the base name of a Windows or Unix path.
    Note that Windows paths can mix forward and backward slashes.
    """
    index = max(path.rfind('/'), path.rfind('\\'))
    if index >= 0:
        return path[index+1:]
    return path

@cxobjc.method(CXUtils, 'dirname_')
def dirname(path):
    """Get the directory name of a Windows or Unix path.
    Note that Windows paths can mix forward and backward slashes.
    """
    index = max(path.rfind('/'), path.rfind('\\'))
    if index >= 0:
        return path[:index].rstrip('/\\')
    return ''


_regex_specials = set('.^$*+?[]|()')

def unescape_regex(string):
    """If a regex can match multiple strings, return None. If it can match only
    one string, this function MAY return that string."""
    in_curly_brace = False
    escape = False
    result = []
    string = string.lstrip('^').rstrip('$')
    for ch in string:
        if ch == '{' and not escape:
            in_curly_brace = True
        elif in_curly_brace:
            if ch == '}':
                return None
            elif not ch.isdigit() and ch != ',':
                in_curly_brace = False

        if escape:
            if ch.isalnum():
                # An escaped letter or number might have special meaning
                return None
            result.append(ch)
            escape = False
        elif ch in _regex_specials:
            return None
        elif ch == '\\':
            escape = True
        else:
            result.append(ch)
    return ''.join(result)


_UNIXVAR_RE = None
def expand_unix_string(environ, string):
    """Expand references to environment variables of the form '${VARNAME}' in
    the specified string.

    Note that unlike os.path.expandvars(), if a given environment variable is
    not set it expands to an empty string.
    """
    if string is None:
        return None
    # pylint: disable-msg=W0603
    global _UNIXVAR_RE
    if not _UNIXVAR_RE:
        _UNIXVAR_RE = re.compile('\${(\w+)}')
    i = 0
    expanded = ''
    for match in _UNIXVAR_RE.finditer(string):
        if match.group(1) in environ:
            expanded += string[i:match.start()] + environ[match.group(1)]
        else:
            expanded += string[i:match.start()]
        i = match.end()
    return expanded + string[i:]


#####
#
# Locale functions
#
#####

def get_preferred_languages(language=None):
    """Returns a sequence of the user's preferred languages if language is
    None; and a sequence of variants otherwise.

    In the None case the result corresponds to the user's preferences on Mac
    OS X, while on Unix it is derived from the user's locale.
    When starting from a single language, the variants added allow broader
    matches to be established. For instance 'pt_br' gives ('pt-br', 'pt', '').
    """

    if not language:
        languages = distversion.get_user_languages()

        if not languages:
            # Theoretically, we should arrange to call
            # locale.setlocale(locale.LC_ALL, '') at program start and use
            # getlocale here. Other functions, e.g. localization, might need
            # the current locale to be set. If we do that, we should use
            # getlocale() here, but for now getdefaultlocale() works fine.
            language, _encoding = locale.getdefaultlocale()
            if language is not None:
                languages = (language,)
            else:
                languages = ()
    else:
        languages = (language,)

    # Add the user's preferred languages
    seen = set(('C', ''))  # Ignore these two 'languages'
    preferred = []
    for language in languages:
        language = language.lower()
        # Mac has the portuguese locales backwards
        if distversion.IS_MACOSX and language == 'pt':
            language= 'pt_br'
        if language in seen:
            continue
        preferred.append(language)
        seen.add(language)

        # Try the fallback
        # We intermingle them with the user's preferences because when the user
        # says he wants ('en_US, 'fr') he really expects us to give him English
        # strings which means trying ('en_US', 'en', 'fr').
        language = language.split('_', 1)[0]
        if language in seen:
            continue
        preferred.append(language)
        seen.add(language)

    # Add the ultimate fallback
    preferred.append('')
    return preferred


@cxobjc.method(CXUtils, 'getLanguageValue_forLanguage_')
def get_language_value(dictionary, language=None):
    """Returns the appropriate value for the specified language from a
    dictionary mapping language names to values.

    The given dictionary should contain a '' key for the default value and may
    contain other keys named like 'pt' or 'pt-br' for language-specific
    values. Empty values are ignored.

    This function will always return a non-empty value if there is at least one
    in the dictonary.
    """
    languages = get_preferred_languages(language)
    for lang in languages:
        if lang in dictionary and dictionary[lang]:
            return (lang, dictionary[lang])

    if dictionary:
        # The dictionary is not empty, but there is no default. Just pick something, predictably.
        languages = dictionary.keys()
        languages.sort()
        for lang in languages:
            if dictionary[lang]:
                return (lang, dictionary[lang])
    return ('', '')


_gettext_translator = None

def cxgettext( message ):
    # pylint: disable-msg=W0603
    global _gettext_translator
    if not _gettext_translator:
        lang = get_preferred_languages()
        locale_path = os.path.join(CX_ROOT, "share", "locale")
        cxlog.log("locale path %s" % cxlog.debug_str(locale_path))
        gettext.bindtextdomain("crossover", locale_path)
        gettext.bind_textdomain_codeset("crossover", "UTF-8")
        gettext.textdomain("crossover")
        _gettext_translator = gettext.translation("crossover", locale_path, languages=lang, fallback=True)
    return _gettext_translator.ugettext(message)


#####
#
# Dealing with subprocesses
#
#####

GRAB = 1
NULL = 2
STDOUT = 3
NOLOG = 4

def run(args, bufsize=0, executable=None, stdout=None, stderr=None,
        preexec_fn=None, shell=False, background=False, cwd=None, env=None,
        universal_newlines=True, logprefix=""):
    """Works the same way as subprocess.Popen, but simplifies retrieving the
    output and exit code. Returns (retcode, out, err) triplets.

    If stdout or stderr are set to GRAB, then they are captured and returned.
    If they are set to NULL, then they are redirected to /dev/null instead.
    If stderr is set to STDOUT, then it is redirected to stdout.
    If the command cannot be run, then returns -1 as the exit code instead of
    raising an exception.

    Finally, the execution is timed and traced with cxlog. If stdout or stderr
    are being captured, then they are logged too, but that can be disabled by
    or-ing NOLOG. If logprefix is set, then the trace is prefixed with it.
    """
    if cxlog.is_on():
        start = time.time()
        if logprefix != "":
            logprefix = logprefix + ": "
        bg = ""
        if background:
            bg = " &"
        cxlog.log("%sRunning: %s%s" % (logprefix, " ".join(cxlog.debug_str(arg) for arg in args), bg))

    out_log = False
    if stdout:
        if (stdout & ~NOLOG) == GRAB and not background:
            subout = subprocess.PIPE
            if stdout == GRAB:
                out_log = True
        elif stdout == NULL:
            subout = open("/dev/null", "w")
        else:
            raise TypeError("invalid value for stdout")
    else:
        subout = None

    err_log = False
    if stderr:
        if (stderr & ~NOLOG) == GRAB and not background:
            suberr = subprocess.PIPE
            if stderr == GRAB:
                err_log = True
        elif stderr == STDOUT:
            suberr = subprocess.STDOUT
        elif stderr == NULL:
            if cxlog.is_on():
                suberr = None
            else:
                suberr = open("/dev/null", "w")
        else:
            raise TypeError("invalid value for stderr")
    else:
        suberr = None

    try:
        subp = subprocess.Popen(args, bufsize, executable, None, subout,
                                suberr, preexec_fn, close_fds=True,
                                shell=shell, cwd=cwd, env=env,
                                universal_newlines=universal_newlines)
        if not background:
            out, err = subp.communicate()
            retcode = subp.wait()
        else:
            out = err = ""
            retcode = 0
    except OSError, exception:
        retcode = -1
        out = ""
        err = cxlog.to_str(exception)

    if isinstance(subout, file):
        # pylint: disable-msg=E0012,E1103
        subout.close()
    if isinstance(suberr, file):
        # pylint: disable-msg=E0012,E1103
        suberr.close()
    if not background and cxlog.is_on():
        cxlog.log("%s%s -> rc=%d  (took %0.3f seconds)" % (logprefix, cxlog.debug_str(os.path.basename(args[0])), retcode, time.time() - start))
        if out_log:
            count = out.count("\n")
            if count > 20:
                cxlog.log("out=<%d lines>" % count)
            else:
                cxlog.log("out=[%s]" % out)
        if err_log:
            count = err.count("\n")
            if count > 20:
                cxlog.log("err=<%d lines>" % count)
            else:
                cxlog.log("err=[%s]" % err)

    return (retcode, out, err)

def system(args, executable=None, preexec_fn=None, shell=False,
           background=False, cwd=None, env=None, logprefix=""):
    """Works the same way as run(), except that neither stdout nor stderr are
    captured, and only the retcode is returned.
    """
    retcode, _out, _err = run(args, executable=executable,
                              preexec_fn=preexec_fn, shell=shell,
                              background=background, cwd=cwd, env=env,
                              logprefix=logprefix)
    return retcode


def launch_url(url):
    """Uses launchurl to open the specified URL in a web browser."""
    return system([os.path.join(CX_ROOT, "bin", "launchurl"), url],
                  background=True)



#####
#
# Select wrapper
#
#####


# Functions and module imports are stored in the same namespace. This means a
# simple 'import select' would get overriden by our select() function, thus
# preventing it from accessing the content of the module, and necessitating
# annoying workarounds (and rightful Pylint warnings).
#
# So we just import the parts we need, renaming them in the process to avoid
# any conflict.
from select import select as select_select
from select import error as select_error

def select(iwtd, owtd, ewtd, timeout=None):
    """Same as select.select() but retries on EINTR (with the same timeout)."""
    while True:
        try:
            return select_select(iwtd, owtd, ewtd, timeout)
        except select_error, err:
            if err.args[0] == errno.EINTR:
                if timeout is not None:
                    timeout = 0
                continue
            else:
                raise



#####
#
# User directories
#
#####

def _xdg_dirs_skip_whitespace(pos, string):
    while pos < len(string) and string[pos] in ' \t':
        pos += 1
    return pos

def _internal_load_xdg_dirs():
    try:
        # pylint: disable-msg=F0401
        # FIXME: Should we just ship this on Linux? It's small, self-contained, and LGPL.
        import xdg.BaseDirectory
        config_file = xdg.BaseDirectory.load_first_config('user-dirs.dirs')
    except:
        cxlog.warn("Cannot load xdg module:\n%s" % traceback.format_exc())
        config_file = None
    if config_file is None:
        config_file = os.path.join(os.environ.get('HOME', '/'), ".config", "user-dirs.dirs")
    result = {}
    try:
        f = open(config_file, 'U')
    except:
        return {}
    try:
        for line in f:
            line = line.rstrip('\n')
            # Based on MIT code in xdg-user-dir-lookup.c in the xdg-user-dirs package.
            pos = _xdg_dirs_skip_whitespace(0, line)

            if line[pos:pos+4] != 'XDG_':
                continue
            pos += 4

            dir_pos = line.find('_DIR', pos)
            if dir_pos == -1:
                continue
            name = line[pos:dir_pos]
            pos = dir_pos + 4

            pos = _xdg_dirs_skip_whitespace(pos, line)

            if line[pos:pos+1] != '=':
                continue
            pos += 1

            pos = _xdg_dirs_skip_whitespace(pos, line)

            if line[pos:pos+1] != '"':
                continue
            pos += 1

            if line[pos:pos+6] == '$HOME/':
                pos += 6
                value_base = os.environ.get('HOME', '/') + '/'
            elif line[pos:pos+1] == '/':
                value_base = ''
            else:
                continue

            value_end = line.find('"', pos)

            if value_end != -1:
                value = value_base + line[pos:value_end].replace('\\', '')
            else:
                # Yes, the standard parser does not mind if there's no terminating quote.
                value = value_base + line[pos:].replace('\\', '')

            result[name] = value
    finally:
        f.close()
    return result

def _load_xdg_dirs():
    global _xdg_dirs
    try:
        return _xdg_dirs
    except NameError:
        pass
    _xdg_dirs = _internal_load_xdg_dirs()
    return _xdg_dirs

def get_download_dir():
    if distversion.IS_MACOSX:
        # FIXME: OSX-specific code here.
        pass
    else:
        xdg_dirs = _load_xdg_dirs()
        if 'DOWNLOAD' in xdg_dirs:
            return xdg_dirs['DOWNLOAD']
        # FIXME: Localize the default somehow?
    return os.path.join(os.environ.get('HOME', '/'), 'Downloads')

def get_desktop_dir():
    if distversion.IS_MACOSX:
        # FIXME: OSX-specific code here.
        pass
    else:
        xdg_dirs = _load_xdg_dirs()
        if 'DESKTOP' in xdg_dirs:
            return xdg_dirs['DESKTOP']
        # FIXME: Localize the default somehow?
    return os.path.join(os.environ.get('HOME', '/'), 'Desktop')



#####
#
# pylint utilities
#
#####

def not_yet_implemented():
    """If a method raises NotImplementedError(), pylint regards it as an
    "abstract method" that must be implemented in all subclasses, and the
    class as an "abstract class". Sometimes this is desired, but in some cases
    we use this exception for parts of the API that we are not using now but
    plan to implement and use in the future. Those functions should instead
    raise cxutils.not_yet_implemented()."""
    return NotImplementedError()


#####
#
# umask cache
#
#####

# Retrieve the umask on startup, that is before we have multiple threads all
# trying to get it at the same time. os.umask() MUST NOT be used in anywhere
# else. Should we need to modify the umask, then we will need to write a
# proper umask() wrapper.
UMASK = os.umask(0)
os.umask(UMASK)
