# (c) Copyright 2009-2010. CodeWeavers, Inc.

import os

import cxlog
import distversion

import tempfile
import urllib2
import cxobjc

import cxutils
import bottlequery

# for localization
from cxutils import cxgettext as _


class DemoUtils(cxobjc.Proxy):
    pass

@cxobjc.method(DemoUtils, 'demoStatus')
def demo_status():
    """Checks the demo-status of the install.  Returns a 2-tuple.
       The first item is a simple boolean:  True if we're in demo
       mode, False if we aren't.  The second item is a string which
       will describe the account used to register the demo, if any.
    """

    winewrapperpath = os.path.join(cxutils.CX_ROOT, "lib", "wine", "winewrapper.exe.so")
    wrapperFile = open(winewrapperpath)
    wrapperBuffer = wrapperFile.read()

    if wrapperBuffer.find("DistinctiveDemoIdentifier") != -1:
        isDemo = True
    else:
        isDemo = False


    if not isDemo:
        registeredTo = "RegisteredTo:"
        brandIndex = wrapperBuffer.find(registeredTo)
        if brandIndex == -1:
            regString = None
        else:
            # Note:  This introduces an obscure corner case.  If the brand is so long
            #   that there are no X's at the end, we'll fail and report no branding.
            endIndex = wrapperBuffer[brandIndex:].find("XXXX")
            if endIndex == -1:
                regString = None
            else:
                regString = wrapperBuffer[brandIndex + len(registeredTo): brandIndex + endIndex]
    else:
        regString = None

    return (isDemo, regString)


@cxobjc.method(DemoUtils, 'canRegister')
def can_register_as_current_user():
    """ Returns True if we have current permissiosn to register.  Return
        False if we need to authenticate as root before registering. 
    """

    wrapperPath = os.path.join(cxutils.CX_ROOT, "lib", "wine")
    return os.access(wrapperPath, os.W_OK)


@cxobjc.method(DemoUtils, 'removeRegMenu')
def remove_registration_menu():
    if not os.path.exists(os.path.join(cxutils.CX_ROOT, 'share', 'crossover', 'data', 'crossover.menu')):
        return 0

    # Remove the 'register and unlock' menu item.
    args = [os.path.join(cxutils.CX_ROOT, "bin", "cxmenu"),
            "--crossover", "--mode", "ignore", "--uninstall",
            "--filter", "StartMenu/Register and Unlock this Demo"]
    menuRemovalResults = cxutils.run(args, stderr=cxutils.GRAB)
    if menuRemovalResults[0] or menuRemovalResults[2]:
        cxlog.log("Warning:  Failed to remove the registration menu item: %s" % menuRemovalResults[2])

    return menuRemovalResults[0]


@cxobjc.method(DemoUtils, 'verifyWinewrapper')
def verify_working_winewrapper():
    """Verify that the current winewrapper actually works.   Return True on success.
    """

    testBottleName = bottlequery.unique_bottle_name(bottlequery.registrationBottlePrefix)

    args = [os.path.join(cxutils.CX_ROOT, "bin", "cxbottle"), "--create", "--bottle", testBottleName]
    cxutils.run(args, stderr=cxutils.GRAB)

    args = [os.path.join(cxutils.CX_ROOT, "bin", "wine"), "--bottle", testBottleName, "--check", "--no-convert", "--wl-app", "winepath.exe", "--unix", "--", "c:/"]
    pathTestResults = cxutils.run(args, stdout=cxutils.GRAB, stderr=cxutils.GRAB)

    args = [os.path.join(cxutils.CX_ROOT, "bin", "cxbottle"), "--delete", "--bottle", testBottleName, "--force"]
    cxutils.run(args)

    if pathTestResults[0] or pathTestResults[1].find("/dosdevices/c:/") == -1:
        cxlog.log("Upgraded winewrapper failed: %s" % pathTestResults[2])
        return False

    return True


@cxobjc.method(DemoUtils, 'regKeyUrlForLogin_andPassword_')
def reg_key_url(login, password):
    fullKeyName = distversion.BUILTIN_PRODUCT_ID + "_" + distversion.PLATFORM + "_" + distversion.CX_VERSION
    fullKeyName = fullKeyName.replace(" ", "_").lower()
    # Truncate the key at the last dot.  That lets us match keys for different subsubversions.
    lastDotIndex = fullKeyName.rfind(".")
    keyName = fullKeyName[0:lastDotIndex]

    # Get an md5 of the password.
    md5hasher = cxutils.md5_hasher()
    md5hasher.update(password)
    mdPassword = md5hasher.hexdigest()

    URL = "HTTP://www.codeweavers.com/bin/register/%s;%s;%s;%s;md5;%s" % (login, mdPassword, keyName, distversion.DEMO_EXTREP, distversion.DEMO_SKU)
    return URL


@cxobjc.method(DemoUtils, 'downloadWrapperWithLogin_andPassword_')
def download_and_brand_winewrapper(login, password):

    """ Attempt to download, brand, and save an enabled winewrapper.
        Returns filename, errorstring where filename will be None if
        the download failed.
    """

    errorString = None
    upgraderFilename = None

    URL = reg_key_url(login, password)

    lastResort = _("If this problem continues, you can also upgrade by downloading and installing an unlocked version.")
    if distversion.HAS_GAMES and not distversion.HAS_OFFICE:
        lastResort = _("If this problem continues, contact CodeWeavers sales support at sales@CodeWeavers.com.")

    while True:

        try:
            wrapperFileLike = urllib2.urlopen(URL)
            wrapperBuffer = wrapperFileLike.read()
        except urllib2.URLError:
            errorString = _("Download failed.\n\n%(lastresort)s") % {'lastresort': lastResort}
            break

        if wrapperBuffer.find("FILE NOT FOUND", 0, 256) != -1:
            errorString = _("Unable to locate an upgrader for %(product)s.\n\n%(lastresort)s") % {'product': distversion.PRODUCT_NAME, 'lastresort': lastResort}
            break
        elif wrapperBuffer.find("USER NOT FOUND", 0, 256) != -1 or wrapperBuffer.find("BAD PASSWORD", 0, 256) != -1:
            errorString = _("Unable to validate your login and password.\n\nVisit store.codeweavers.com to make sure you have a valid login, and try again.\n\n%(lastresort)s") % {'lastresort': lastResort}
            break
        elif wrapperBuffer.find("USER EXPIRED", 0, 256) != -1:
            errorString = _("Your support period has expired. You are no longer entitled to product upgrades.\n\nVisit store.codeweavers.com to extend your support contract.")
            break
        elif wrapperBuffer.find("ERROR", 0, 256) != -1:
            errorString = _("Download failed.\n\n%(lastresort)s") % {'lastresort': lastResort}
            break


        # brand the key while it's still in memory.
        XString = "X" * len(login)
        matchString = "RegisteredTo:" + XString
        brand = "RegisteredTo:" + login
        wrapperBuffer = wrapperBuffer.replace(matchString, brand, 1)

        # And, write it out.
        winewrapperdir = os.path.join(cxutils.CX_ROOT, "lib", "wine")
        upgraderFileno, upgraderFilename = tempfile.mkstemp(dir=winewrapperdir)
        outfile = os.fdopen(upgraderFileno, "w")
        outfile.write(wrapperBuffer)
        cxlog.log("Key downloaded and written to %s" % upgraderFilename)
        outfile.close()

        break


    return upgraderFilename, errorString


