# (c) Copyright 2009-2010. CodeWeavers, Inc.

import re
import sys

import cxlog
import cxutils


def not_much(*_args):
    pass

class error(Exception):
    pass

class MissingDiskError(error):
    pass

class Volume(object):
    device = ''
    # A string that uniquely identifies this device and does not change during
    # its lifetime. If possible, this is the filename of a block device.

    mounted = False
    # True if the device is currently mounted. This can lag behind the actual
    # device state, in which case calling mount, remount, or unmount will update
    # it.

    mountpoint = ''
    # If the device is mounted, this is its current mount point. This can change
    # if the device is mounted multiple times.

    label = ''
    # If possible, this is the label of the drive. An empty string otherwise.

    is_disc = False
    # True if the device is a CD or DVD.

    has_disk = True
    # True if a disk is present, or if it's unknown whether a disk is present.

    def mount(self):
        """Mount the drive if it isn't mounted. Raises MissingDiskError if there
        is no disk in the drive. Other exceptions are also possible."""
        raise cxutils.not_yet_implemented()

    def remount(self, _options='', _filesystem=''):
        """Remount the drive to change the options and/or filesystem. Raises
        MissingDiskError if there is no disk in the drive. Other exceptions are
        also possible."""
        raise cxutils.not_yet_implemented()

    def unmount(self):
        """Unmount the drive if it's mounted. This can raise an exception if it
        fails."""
        raise cxutils.not_yet_implemented()

class HALVolume(Volume):
    udi = ''
    # The HAL unique device id

class HALMountPoints(object):
    def new_device(self, udi, *_rest):
        #  Get the device object
        device_obj = self.bus.get_object('org.freedesktop.Hal', udi)

        #  Get an interface to the device
        device = self.dbus.Interface(device_obj, 'org.freedesktop.Hal.Device')

        if not device.QueryCapability('volume'):
            return

        mp_path = device.GetProperty('volume.mount_point')
        if mp_path != "":
            mountpoint = HALVolume()
            mountpoint.udi = udi
            mountpoint.device = device.GetProperty('block.device')
            mountpoint.mountpoint = mp_path
            mountpoint.label = device.GetProperty('volume.label')
            if device.GetProperty('volume.is_disc'):
                mountpoint.is_disc = True
            else:
                mountpoint.is_disc = False
            self.devices[udi] = mountpoint
            self.add_notify(mountpoint)

        if udi in self.device_signals:
            return

        def on_property_modified(*_args):
            if device.GetProperty('volume.mount_point') != "":
                self.new_device(udi)
            else:
                self.unmounted_device(udi)

        self.device_signals[udi] = device.connect_to_signal("PropertyModified", on_property_modified)

    def unmounted_device(self, udi):
        if udi in self.devices:
            self.del_notify(self.devices[udi])
            del self.devices[udi]

    def removed_device(self, udi, *_rest):
        if udi in self.device_signals:
            self.unmounted_device(udi)
            self.device_signals[udi].remove()
            del self.device_signals[udi]

    def refresh(self):
        # find all mountable devices
        udis = self.hal.FindDeviceByCapability('volume')
        for udi in udis:
            self.new_device(udi)

    def __init__(self, add_notify=not_much, del_notify=not_much):
        self.device_signals = {}
        self.devices = {}

        self.signals = []

        try:
            import dbus
            self.dbus = dbus
            # Initialize the glib main loop
            import dbus.glib
        except ImportError:
            cxlog.warn("unable to import dbus")
            sys.stderr.write("\nYou may be missing the Python D-Bus packages. If this is the case you may try running one of the following commands as root:\n")
            sys.stderr.write("Debian/Ubuntu: apt-get install python-dbus\n")
            sys.stderr.write("Fedora:        yum install dbus-python\n")
            sys.stderr.write("Mandriva:      urpmi dbus-python\n")
            sys.stderr.write("SUSE:          zypper install dbus-1-python\n")

            # We cannot do anything without the above modules
            raise

        # get a connection to the system bus
        self.bus = dbus.SystemBus()

        # get a HAL object and an interface to HAL
        self.hal_obj = self.bus.get_object('org.freedesktop.Hal','/org/freedesktop/Hal/Manager')
        self.hal = dbus.Interface (self.hal_obj, 'org.freedesktop.Hal.Manager')

        if not self.hal:
            raise Exception("Cannot connect to HAL")

        self.add_notify = add_notify
        self.del_notify = del_notify

        self.signals.append(self.hal.connect_to_signal("DeviceAdded", self.new_device))
        self.signals.append(self.hal.connect_to_signal("DeviceRemoved", self.removed_device))
        self.signals.append(self.hal.connect_to_signal("NewCapability", self.new_device))

        self.refresh()

    def close(self):
        self.add_notify = not_much
        self.del_notify = not_much
        for signal in self.signals:
            signal.remove()
        for signal in self.device_signals.values():
            signal.remove()

    def __del__(self):
        self.close()

class UnixVolume(Volume):
    # A mount point obtained from the mount command or /etc/fstab.
    pass

class FailsafeMountPoints(object):

    # The Linux mtab format assumes device paths don't contain spaces.
    # Otherwise it is ambiguous.
    LINUX_MTAB_RE = re.compile('(?P<device>\S+)\s+(?P<mountpoint>.+)\s+(?P<fs>\w+)\s+\S+\s+\d+\s+\d+$')
    SOLARIS_MTAB_RE = re.compile('(?P<device>.+)\t(?P<mountpoint>.+)\t(?P<fs>\w+)\t')
    LINUX_MOUNT_RE = re.compile('(?P<device>.+) on (?P<mountpoint>.+) type (?P<fs>\w+) \(.*\)')
    FREEBSD_MOUNT_RE = re.compile('(?P<device>.+) on (?P<mountpoint>.+) \((?P<fs>\w+)(?:,.*)?\)')

    def check_mount_cmd(self):
        mountdata = None
        for filename in ('/etc/mtab', '/etc/mnttab'):
            try:
                thisfile = open(filename, 'r')
                mountdata = list(thisfile.readlines())
                thisfile.close()
                regexps = (self.LINUX_MTAB_RE, self.SOLARIS_MTAB_RE)
                break
            except IOError, ioerror:
                cxlog.log('could not read %s: %s' % (filename, ioerror))

        if mountdata is None:
            retcode, out, err = cxutils.run(('mount', ), stdout=cxutils.GRAB,
                                            stderr=cxutils.GRAB)
            if retcode:
                cxlog.log("mount failed (%s):\n%s%s" % (retcode, out, err))
            mountdata = out.split('\n')
            regexps = (self.LINUX_MOUNT_RE, self.FREEBSD_MOUNT_RE)

        known_devices = self.devices.keys()

        # Given the call parameters, communicate() always returns a string
        for line in mountdata:
            if line == '':
                continue

            for regexp in regexps:
                match = regexp.match(line)
                if match:
                    break
            if match is None:
                continue

            device = match.group('device')
            mp_path = match.group('mountpoint')
            filesystem = match.group('fs')

            if not device.startswith('/') or not mp_path.startswith('/') or \
                filesystem == 'sysfs':
                continue

            if device in known_devices:
                known_devices.remove(device)
            else:
                mountpoint = UnixVolume()
                mountpoint.device = device
                mountpoint.mountpoint = mp_path
                if filesystem in set(('cd9660',   # FreeBSD
                                      'iso9660',  # Linux
                                      'hsfs',     # Solaris
                                      'udf',      # Linux
                                      )):
                    mountpoint.is_disc = True
                else:
                    mountpoint.is_disc = False
                self.devices[device] = mountpoint
                self.add_notify(mountpoint)

        for device in known_devices:
            self.del_notify(self.devices[device])
            del self.devices[device]

        return True # continue timer

    def __init__(self, add_notify=not_much, del_notify=not_much):
        # No Hal, so we resort to old-school methods.
        import gobject

        self.devices = {}

        self.add_notify = add_notify
        self.del_notify = del_notify

        self.timer_src = None

        self.check_mount_cmd()

        self.timer_src = gobject.timeout_add(2000, self.check_mount_cmd)

    def close(self):
        if self.timer_src is not None:
            import gobject
            gobject.source_remove(self.timer_src)
            self.timer_src = None

    def __del__(self):
        self.close()

# A list of notifier classes, with "better" notifiers at the start
notifier_classes = [
    HALMountPoints,
    FailsafeMountPoints,
    ]

class MountPointsNotifier(object):
    "A class that returns the combined results from the full set of notifiers"
    def __init__(self, add_notify=not_much, del_notify=not_much):
        self.add_notify = add_notify
        self.del_notify = del_notify

        self.notifiers = []
        self.volumes = [{}]
        for notifier_class in notifier_classes:
            try:
                def on_add(volume, index=len(self.notifiers)):
                    self.on_add(volume, index)
                def on_del(volume, index=len(self.notifiers)):
                    self.on_del(volume, index)
                notifier = notifier_class(on_add, on_del)
                self.notifiers.append(notifier)
                self.volumes.append({})
            except:
                import traceback
                cxlog.log("unable to create the %s object:\n%s" % (notifier_class.__name__, traceback.format_exc()))

        if not self.notifiers:
            cxlog.warn('could not create any mount point notifier')

        self.volumes.pop()

    def on_add(self, volume, index):
        self.volumes[index][volume.device] = volume

        for i in range(len(self.notifiers)):
            if volume.device in self.volumes[i]:
                if i < index:
                    # We already have a volume for this device, and it's better
                    return
                elif i == index:
                    continue
                else:
                    # This new device is better than the current best
                    self.del_notify(self.volumes[i][volume.device])
                    break

        self.add_notify(volume)

    def on_del(self, volume, index):
        del self.volumes[index][volume.device]
        self.del_notify(volume)

        for volumes in self.volumes:
            if volume.device in volumes:
                self.add_notify(volumes[volume.device])
                break

    def close(self):
        self.add_notify = not_much
        self.del_notify = not_much
        for notifier in self.notifiers:
            notifier.close()

    def _get_devices(self):
        devices_dict = {}
        for volumes in self.volumes[::-1]:
            devices_dict.update(volumes)
        return devices_dict

    devices = property(_get_devices)

def GetVolumeList():
    mountpointsnotify = MountPointsNotifier()
    result = mountpointsnotify.devices.values()
    mountpointsnotify.close()
    return result

